:root {
    --bk-color: #FFFFFF;
    --blk-color: #f8f8f8;

    --fg-color:  #E0DEE0;
    --scroll-bar: #71c9d3;

}

body {
    background-color: var(--bk-color);
    overflow: hidden;
    padding: 0px;
    margin: 0px;
}

svg {
    background-color: var(--bk-color);
}

p {
    padding: 0px;
    margin: 0px;
}

.title {
    background-color: var(--fg-color);
    font-family: 'Open Sans', sans-serif, Impact, Geneva ;
    font-size: 38px;
    color: black;
    line-height: 50px;
    padding: 10px;
    width: 100%;
    box-shadow: 0 4px 6px rgba(0, 0, 0, .26);
}

.titleWrapper {
    height: 5px;
    width: 100%;
}

.labels {
    font-size: 13px;
    font-family: cursive;
}

.tempclass {
    font-size: 1px;
}

.shadow_box{
    /* border: 2px solid var(--fg-color); */
    box-shadow: 0 4px 4px 0 rgba(0, 0, 0, 0.16), 0 0 0 2px rgba(0, 0, 0, 0.08);
    border-radius: 2px;
    align-content: center;
    /* width: 15vw; */
    width:250px;
    font-family: Arial, Helvetica, sans-serif;
    position: absolute;
    left: 5px;
    background-color: var(--blk-color);
    color: gray;
}
#import{
    top:80px;
    height: 80px;
}
#tips{
    bottom: 5px
}
h4 {
    padding:2px 7px;
    margin:7px 0px;
}
#layerInfo {
    top: 170px;
}

#layerName {
    font-size: 0.9em;
    padding: 5px 7px;
    border-radius: 2px 2px 0px 0px;
    word-wrap: break-word;
}

#layerConfig {
    padding: 2px 7px;
    border-radius: 0px 0px 2px 2px;
    font-size: 0.7em;
    max-height: 60vh;
    overflow-y: auto;
    border-top: 1px solid rgba(0, 0, 0, 0.3);
    word-wrap: break-word;
}

/* toggle switch */
#order{
    padding:7px;
}
.switch {
    position: relative;
    display: inline-block;
    width: 55px;
    height: 20px;
    /* padding:2px 7px; */
    margin:0px 7px 2px 7px;
  }

  .switch input {display:none;}

  .slider {
    position: absolute;
    cursor: pointer;
    top: 0;
    left: 0;
    right: 0;
    bottom: 0;
    background-color: #ccc;
    -webkit-transition: .4s;
    transition: .4s;
  }

  .slider:before {
    position: absolute;
    content: "";
    height: 16px;
    width: 16px;
    left: 2px;
    bottom: 2px;
    background-color: white;
    -webkit-transition: .4s;
    transition: .4s;
  }

  input:checked + .slider {
    background-color: var(--fg-color);
  }

  input:focus + .slider {
    box-shadow: 0 0 1px var(--fg-color);
  }

  input:checked + .slider:before {
    -webkit-transform: translateX(35px);
    -ms-transform: translateX(35px);
    transform: translateX(35px);
  }

  /* Rounded sliders */
  .slider.round {
    border-radius: 17px;
  }

  .slider.round:before {
    border-radius: 50%;
  }
  .slider:after
  {
   content:'TB';
   color: white;
   display: block;
   position: absolute;
   transform: translate(-50%,-50%);
   top: 50%;
   left: 50%;
   font-size: 10px;
   font-family: Verdana, sans-serif;
  }

  input:checked + .slider:after
  {
    content:'BT';
  }

/* the scroll bar */
/* Use -webkit- only if supporting: Chrome < 54, iOS < 9.3, Android < 4.4.4 */
/* .shadow {

    -webkit-filter: drop-shadow( -5px -5px 5px #000);
    filter: drop-shadow( -5px -5px 5px #000);
}*/

/*scroll bar*/

::-webkit-scrollbar {
    width: 6px;
}

/* Track */

::-webkit-scrollbar-track {
    -webkit-box-shadow: inset 0 0 3px rgba(0, 0, 0, 0.3);
    -webkit-border-radius: 6px;
    border-radius: 3px;
}

/* Handle */

::-webkit-scrollbar-thumb {
    -webkit-border-radius: 6px;
    border-radius: 3px;
    background: var(--scroll-bar);
    opacity: 0.5;
    -webkit-box-shadow: inset 0 0 3px rgba(0, 0, 0, 0.25);
}

/* ::-webkit-scrollbar-thumb:window-inactive {
    background:var(--scroll-bar);
} */


